@@ -15,6 +15,7 @@ import struct
 import time
 from contextlib import suppress
 from io import StringIO
+from subprocess import TimeoutExpired
 from typing import Any, Callable, Dict, List, Optional, Tuple
 
 import configobj
@@ -81,15 +82,19 @@ class NoDHCPLeaseMissingDhclientError(NoDHCPLeaseError):
     """Raised when unable to find dhclient."""
 
 
-def maybe_perform_dhcp_discovery(distro, nic=None, dhcp_log_func=None):
+def maybe_perform_dhcp_discovery(
+    distro,
+    nic=None,
+    dhcp_log_func: Optional[Callable[[str, str, str], None]] = None,
+) -> Dict[str, Any]:
     """Perform dhcp discovery if nic valid and dhclient command exists.
 
     If the nic is invalid or undiscoverable or dhclient command is not found,
     skip dhcp_discovery and return an empty dict.
 
     @param nic: Name of the network interface we want to run dhclient on.
-    @param dhcp_log_func: A callable accepting the dhclient output and error
-        streams.
+    @param dhcp_log_func: Callable accepting the interface and client's
+        stdout, stderr streams.
     @return: A list of dicts representing dhcp options for each lease obtained
         from the dhclient discovery if run, otherwise an empty list is
         returned.
@@ -122,7 +127,7 @@ def networkd_load_leases(leases_d=None):
     if leases_d is None:
         leases_d = NETWORKD_LEASES_DIR
 
-    ret = {}
+    ret: Dict[str, dict] = {}
     if not os.path.isdir(leases_d):
         return ret
     for lfile in os.listdir(leases_d):
@@ -200,7 +205,7 @@ class DhcpClient(abc.ABC):
     def dhcp_discovery(
         self,
         interface: str,
-        dhcp_log_func: Optional[Callable] = None,
+        dhcp_log_func: Optional[Callable[[str, str, str], None]] = None,
         distro=None,
     ) -> Dict[str, Any]:
         """Run dhcp client on the interface without scripts or filesystem
@@ -208,8 +213,8 @@ class DhcpClient(abc.ABC):
 
         @param interface: Name of the network interface on which to send a
             dhcp request
-        @param dhcp_log_func: A callable accepting the client output and
-            error streams.
+        @param dhcp_log_func: Callable accepting the interface and client's
+            stdout, stderr streams.
         @param distro: a distro object for network interface manipulation
         @return: dict of lease options representing the most recent dhcp lease
             parsed from the dhclient.lease file
@@ -234,7 +239,7 @@ class IscDhclient(DhcpClient):
         """
         lease_regex = re.compile(r"lease {(?P<lease>.*?)}\n", re.DOTALL)
         dhcp_leases: List[Dict] = []
-        if len(lease_content) == 0:
+        if not lease_content:
             return []
         for lease in lease_regex.findall(lease_content):
             lease_options = []
@@ -291,15 +296,15 @@ class IscDhclient(DhcpClient):
     def dhcp_discovery(
         self,
         interface: str,
-        dhcp_log_func: Optional[Callable] = None,
+        dhcp_log_func: Optional[Callable[[str, str, str], None]] = None,
         distro=None,
     ) -> Dict[str, Any]:
         """Run dhclient on the interface without scripts/filesystem artifacts.
 
         @param interface: Name of the network interface on which to send a
             dhcp request
-        @param dhcp_log_func: A callable accepting the dhclient output and
-            error streams.
+        @param dhcp_log_func: Callable accepting the interface and client's
+            stdout, stderr streams.
         @param distro: a distro object for network interface manipulation
         @return: dict of lease options representing the most recent dhcp lease
             parsed from the dhclient.lease file
@@ -417,7 +422,7 @@ class IscDhclient(DhcpClient):
                 0.01 * 1000,
             )
         if dhcp_log_func is not None:
-            dhcp_log_func(out, err)
+            dhcp_log_func(interface, out, err)
         lease = self.get_newest_lease(interface)
         if lease:
             return lease
@@ -452,14 +457,27 @@ class IscDhclient(DhcpClient):
         ("0.0.0.0/0", "192.168.128.1")
         ]
 
+        # unknown-121 option format
+        sr3 = parse_static_routes(\
+        "0:a:0:0:1:20:a8:3f:81:10:a:0:0:1:20:a9:fe:a9:fe:a:0:0:1")
+        sr3 = [
+            ("0.0.0.0/0", "10.0.0.1"),
+            ("168.63.129.16/32", "10.0.0.1"),
+            ("169.254.169.254/32", "10.0.0.1"),
+        ]
+
         Python version of isc-dhclient's hooks:
            /etc/dhcp/dhclient-exit-hooks.d/rfc3442-classless-routes
         """
         # raw strings from dhcp lease may end in semi-colon
         rfc3442 = routes.rstrip(";")
-        tokens = [tok for tok in re.split(r"[, .]", rfc3442) if tok]
+        tokens = [tok for tok in re.split(r"[, . :]", rfc3442) if tok]
         static_routes: List[Tuple[str, str]] = []
 
+        # Handle unknown-121 format by converting hex to base 10.
+        if ":" in rfc3442:
+            tokens = [str(int(tok, 16)) for tok in tokens]
+
         def _trunc_error(cidr, required, remain):
             msg = (
                 "RFC3442 string malformed.  Current route has CIDR of %s "
@@ -595,19 +613,20 @@ class IscDhclient(DhcpClient):
 
 class Dhcpcd(DhcpClient):
     client_name = "dhcpcd"
+    timeout = 300
 
     def dhcp_discovery(
         self,
         interface: str,
-        dhcp_log_func: Optional[Callable] = None,
+        dhcp_log_func: Optional[Callable[[str, str, str], None]] = None,
         distro=None,
     ) -> Dict[str, Any]:
         """Run dhcpcd on the interface without scripts/filesystem artifacts.
 
         @param interface: Name of the network interface on which to send a
             dhcp request
-        @param dhcp_log_func: A callable accepting the client output and
-            error streams.
+        @param dhcp_log_func: Callable accepting the interface and client's
+            stdout, stderr streams.
         @param distro: a distro object for network interface manipulation
         @return: dict of lease options representing the most recent dhcp lease
             parsed from the dhclient.lease file
@@ -637,7 +656,7 @@ class Dhcpcd(DhcpClient):
             if is_ib_interface(interface):
                 infiniband_argument = ["--clientid"]
             command = [
-                self.dhcp_client_path,  # pyright: ignore
+                self.client_name,
                 "--ipv4only",  # only attempt configuring ipv4
                 "--waitip",  # wait for ipv4 to be configured
                 "--persistent",  # don't deconfigure when dhcpcd exits
@@ -651,7 +670,7 @@ class Dhcpcd(DhcpClient):
                 timeout=self.timeout,
             )
             if dhcp_log_func is not None:
-                dhcp_log_func(out, err)
+                dhcp_log_func(interface, out, err)
             lease = self.get_newest_lease(interface)
             # Attempt cleanup and leave breadcrumbs if it fails, but return
             # the lease regardless of failure to clean up dhcpcd.
@@ -698,9 +717,17 @@ class Dhcpcd(DhcpClient):
                 return lease
             raise NoDHCPLeaseError("No lease found")
 
+        except TimeoutExpired as error:
+            LOG.debug(
+                "dhcpcd timed out after %s seconds: stderr: %r stdout: %r",
+                error.timeout,
+                error.stderr,
+                error.stdout,
+            )
+            raise NoDHCPLeaseError from error
         except subp.ProcessExecutionError as error:
             LOG.debug(
-                "dhclient exited with code: %s stderr: %r stdout: %r",
+                "dhcpcd exited with code: %s stderr: %r stdout: %r",
                 error.exit_code,
                 error.stderr,
                 error.stdout,
@@ -773,14 +800,29 @@ class Dhcpcd(DhcpClient):
         subnet_cidr='20'
         subnet_mask='255.255.240.0'
         """
+        LOG.debug(
+            "Parsing dhcpcd lease for interface %s: %r", interface, lease_dump
+        )
 
         # create a dict from dhcpcd dump output - remove single quotes
-        lease = dict(
-            [
-                a.split("=")
-                for a in lease_dump.strip().replace("'", "").split("\n")
-            ]
-        )
+        try:
+            lease = dict(
+                [
+                    a.split("=", maxsplit=1)
+                    for a in lease_dump.strip().replace("'", "").split("\n")
+                    if "=" in a
+                ]
+            )
+            if not lease:
+                msg = (
+                    "No valid DHCP lease configuration "
+                    "found in dhcpcd lease: %r"
+                )
+                LOG.error(msg, lease_dump)
+                raise InvalidDHCPLeaseFileError(msg % lease_dump)
+        except ValueError as error:
+            LOG.error("Error parsing dhcpcd lease: %r", lease_dump)
+            raise InvalidDHCPLeaseFileError from error
 
         # this is expected by cloud-init's code
         lease["interface"] = interface
@@ -820,7 +862,7 @@ class Dhcpcd(DhcpClient):
             return self.parse_dhcpcd_lease(
                 subp.subp(
                     [
-                        self.dhcp_client_path,
+                        self.client_name,
                         "--dumplease",
                         "--ipv4only",
                         interface,
@@ -874,19 +916,19 @@ class Udhcpc(DhcpClient):
 
     def __init__(self):
         super().__init__()
-        self.lease_file = None
+        self.lease_file = ""
 
     def dhcp_discovery(
         self,
         interface: str,
-        dhcp_log_func: Optional[Callable] = None,
+        dhcp_log_func: Optional[Callable[[str, str, str], None]] = None,
         distro=None,
     ) -> Dict[str, Any]:
         """Run udhcpc on the interface without scripts or filesystem artifacts.
 
         @param interface: Name of the network interface on which to run udhcpc.
-        @param dhcp_log_func: A callable accepting the udhcpc output and
-            error streams.
+        @param dhcp_log_func: Callable accepting the interface and client's
+            stdout, stderr streams.
         @return: A list of dicts of representing the dhcp leases parsed from
             the udhcpc lease file.
         """
@@ -904,7 +946,7 @@ class Udhcpc(DhcpClient):
         util.write_file(udhcpc_script, UDHCPC_SCRIPT, 0o755)
 
         cmd = [
-            self.dhcp_client_path,
+            self.client_name,
             "-O",
             "staticroutes",
             "-i",
@@ -944,7 +986,7 @@ class Udhcpc(DhcpClient):
             raise NoDHCPLeaseError from error
 
         if dhcp_log_func is not None:
-            dhcp_log_func(out, err)
+            dhcp_log_func(interface, out, err)
 
         return self.get_newest_lease(interface)
 
